/*	Author: A.Voss@FH-Aachen.de
 * 
 * 	Thema move, move assignment, r- und l-value
 */

#include <iostream>
#include <cstdlib>
#include <string>
#include <cmath>
#include <utility>
using namespace std;

#include "../../include/tools.h"

#pragma GCC optimize "-fno-elide-constructors"

class I {
	int* p;
	
	static int* alloc(int n) { ++cntalloc; return new int{n}; }

public:    
	static int cntalloc;
	static int cntobj;

    I(int n) : p(alloc(n)) { 
		tools_log(); cout << "ctor I(" << *p << ")" << endl; 
		++cntobj;
	}
	
	~I() {
		int n = (p) ? *p : -1;
		tools_log(); cout << "dtor I(" << n << ")" << endl; 
		delete p;
		--cntobj;
	}
    
    I(const I & i) : p(alloc(*i.p)) { 
        tools_log(); cout << "ctor I(&)" << endl; 
        ++cntobj;
    }
    
    I & operator=(const I & i) {
		*i.p = *i.p;
        tools_log(); cout << "op=(&)" << endl; return *this; 
    }
     
    I & operator*=(const I & i) {
		*p *= *i.p; 
        return *this; 
    }
    
    friend I operator*(I lhs, const I & rhs) { 
        tools_log(); cout << "op*" << endl; 
        lhs*=rhs; 
        return lhs; 
    }

	#ifdef _WITHOUT_
	
    I(I && i) : p(nullptr) { 
        tools_log(); cout << "ctor I(&&)" << endl; 
        swap(p,i.p);	// den Speicher aus i uebernehmen
        ++cntobj;
    }

    I & operator=(I && i) {
        tools_log(); cout << "op=(&&)" << endl; 
        swap(p,i.p);	// den Speicher aus i uebernehmen
        return *this; 
    }
    
    #endif

};

// statische Felder muessen ausserhalb angelegt und init. werden
int I::cntalloc = 0;
int I::cntobj = 0;

int main()
{
    tools_log(); cout << "start" << endl;

	{
    I i1(1),i2(2),i3(3);

    tools_log(); cout << "i3=i1*ci" << endl;
    i3 = i1*i2;
	}

    tools_log(); cout << "alloc cnt: " << I::cntalloc << endl;
    tools_log(); cout << "obj cnt:   " << I::cntobj << endl;

    return EXIT_SUCCESS;
}

